import numpy as np
from transformers import Wav2Vec2CTCTokenizer, Wav2Vec2FeatureExtractor, Wav2Vec2Processor, Wav2Vec2ForCTC
from transformers import TrainingArguments, Trainer

from pyw2v2.components.metrics import Metrics
from pyw2v2.external.data_collator_ctc import DataCollatorCTCWithPadding


class ModelCTC:

    def __init__(self, config):
        self._model = None
        self._metrics = None
        self._processor = None
        self._data_collator = None
        self._training_args = None

        self._init_processor(config)
        self._init_metrics(config)
        self._init_data_collator()
        self._init_model(config)
        self._init_training_args(config)
        
    @property
    def processor(self):
        return self._processor
    
    @processor.setter
    def processor(self, var):
        self._processor = var

    def _init_processor(self, config):
        config.processor.tokenizer.vocab_file = config.common.vocab_file
        tokenizer = Wav2Vec2CTCTokenizer(**config.processor.tokenizer)
        feature_extractor = Wav2Vec2FeatureExtractor(**config.processor.feature_extractor)

        processor = Wav2Vec2Processor(feature_extractor=feature_extractor, tokenizer=tokenizer)
        processor.save_pretrained(config.common.model_path)
        self._processor = processor

    def _init_metrics(self, config):
        self._metrics = Metrics(*config.common.metrics)

    def _init_data_collator(self):
        self._data_collator = DataCollatorCTCWithPadding(processor=self._processor, padding=True)

    def _init_model(self, config):
        if not config.common.checkpoint_model:
            print(f"Loading pretrained model {config.common.pretrained_model}")
            config.model.pretrained_model_name_or_path = config.common.pretrained_model
            config.model.pad_token_id = self._processor.tokenizer.pad_token_id
            config.model.vocab_size = len(self._processor.tokenizer)
            self._model = Wav2Vec2ForCTC.from_pretrained(**config.model)
        else:
            print(f"Loading from checkpoint {config.common.checkpoint_model}")
            self._model = Wav2Vec2ForCTC.from_pretrained(config.common.checkpoint_model).to("cuda")
    
    def _init_training_args(self, config):
        config.training_args.output_dir = config.common.model_path
        self._training_args = TrainingArguments(**config.training_args)
        
    def _compute_metrics(self, pred):
        pred_logits = pred.predictions
        pred_ids = np.argmax(pred_logits, axis=-1)

        pred.label_ids[pred.label_ids == -100] = self._processor.tokenizer.pad_token_id
        pred_str = self._processor.batch_decode(pred_ids)
        label_str = self._processor.batch_decode(pred.label_ids, group_tokens=False)

        res = self._metrics.compute(predictions=pred_str, references=label_str)
        return res
    
    def train(self, train_set, eval_set):
        trainer = Trainer(
            model=self._model,
            data_collator=self._data_collator,
            args=self._training_args,
            compute_metrics=self._compute_metrics,
            train_dataset=train_set,
            eval_dataset=eval_set,
            tokenizer=self._processor.feature_extractor)
        trainer.train()



import librosa
import torchaudio
import numpy as np
from datasets import Dataset
# from datasets.utils.logging import set_verbosity_error
# set_verbosity_error()


class DatasetPreprocessor:
    
    def __init__(self, config, processor=None):
        self._text_column = config.text_column
        self._path_column = config.path_column
        self._remove_chars = config.remove_chars
        self._replace_chars = config.replace_chars
        self._sampling_rate = config.sampling_rate
        self._num_proc = config.n_jobs
        self._processor = processor
        
    @property
    def processor(self):
        return self._processor
    
    @processor.setter
    def processor(self, var):
        self._processor = var

    def _to_lower(self, batch):
        batch[self._text_column] = batch[self._text_column].lower() 
        return batch

    def _remove_characters(self, batch):
        for char in self._remove_chars:
            batch[self._text_column] = batch[self._text_column].replace(char, '')
        return batch
    
    def _replace_characters(self, batch):
        for k, v in self._replace_chars.items():
            batch[self._text_column] = batch[self._text_column].replace(k, v)
        return batch
    
    def _speech_file_to_array(self, batch):
        speech_array, sampling_rate = torchaudio.load(batch[self._path_column])
        batch["speech"] = speech_array[0].numpy()
        batch["sampling_rate"] = sampling_rate
        batch["target_text"] = batch[self._text_column]
        return batch
    
    def _resample(self, batch):
        batch["speech"] = librosa.resample(np.asarray(batch["speech"]), batch["sampling_rate"], self._sampling_rate)
        batch["sampling_rate"] = self._sampling_rate
        return batch

    def _prepare_batch(self, batch):
        # check that all files have the correct sampling rate
        assert (
            len(set(batch["sampling_rate"])) == 1
        ), f"Make sure all inputs have the same sampling rate of {self._processor.feature_extractor.sampling_rate}."

        batch["input_values"] = self._processor(batch["speech"], sampling_rate=batch["sampling_rate"][0]).input_values
        
        with self._processor.as_target_processor():
            batch["labels"] = self._processor(batch["target_text"]).input_ids
        return batch
    
    @staticmethod
    def dataset_head(dataset: Dataset, n):
        return Dataset.from_dict(dataset[:n])

    def process(self, dataset: Dataset, n_samples=None):
        if n_samples:
            dataset = self.dataset_head(dataset, n_samples)
        dataset = dataset.map(self._remove_characters)
        dataset = dataset.map(self._replace_characters)
        dataset = dataset.map(self._to_lower)
        dataset = dataset.map(self._speech_file_to_array)
        dataset = dataset.map(self._resample, num_proc=self._num_proc)
        dataset = dataset.map(self._prepare_batch, batch_size=8, batched=True, num_proc=self._num_proc)
        return dataset


from setuptools import setup, find_packages

VERSION = '0.0.1a1'
DESCRIPTION = 'Simple wav2vec2 wrapper'
LONG_DESCRIPTION = 'A simple wrapper for wav2vec2 for accelerated ASR research.'

# Setting up
setup(
    name="pyw2v2",
    version=VERSION,
    author="vbrydik (Vitalii Brydinskyi)",
    author_email="<vbrydinskyi@gmail.com>",
    url="https://github.com/vbrydik/pyw2v2",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    packages=find_packages(),
    install_requires=["torch", "torchaudio", "librosa", "transformers", "datasets"],
    keywords=['python'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)